<?php
/**
 * Please see weepie-framework.php for more details.
 */

namespace WpieFw\Hooks;

use WpieFw\Wpie\WpieGlobals;
use WpieFw\Helpers\WpieMiscHelper;
use WpieFw\Helpers\WpieMultisiteHelper;
use WpieFw\Exceptions\WpieExceptionInterface;
use WpieFw\Exceptions\WpieExceptionLogger;

if( ! defined( 'ABSPATH' ) ) exit;

/**
 * WpieHooksFrontend Class
 *
 * @author $Author: Vincent Weber <vincent@webrtistik.nl> $
 * @since 1.4.0
 */
final class WpieHooksFrontend extends WpieHooks
{
	/**
	 * Constructor
	 *
	 * @access public
	 *
	 * @since 1.4.0
	 *
	 * @param WpieGlobals $globals
	 */
	public function __construct( WpieGlobals $globals )
	{
		parent::__construct( $globals );
	}

	/**
	 * {@inheritDoc}
	 * @see WpieHooks::add()
	 */
	public function add()
	{
		// Frontend hooks
		if( !is_admin() ) {
			/**
			 * description for hook activate_{$nameSpace}_do_frontend_hooks
			 */
			if( $this->doFrontend ===  apply_filters( $this->globals->nameSpace . '_do_frontend_hooks', true ) ) {
				add_action( 'wp_enqueue_scripts', [$this, 'setStylesFrontend'] );
				add_action( 'wp_enqueue_scripts', [$this, 'setScriptsFrontend'] );
				add_action( 'wp_head', [$this, 'printScriptsFrontendVars'], 8 );
				add_action( 'wp_head', [$this, 'printScriptsFrontend'] );
				add_action( 'template_redirect', [ $this, 'renderTemplate' ] );

				if( isset( $_REQUEST['action'] ) && 'wpie-footer-action' === $_REQUEST['action'] ) {
					add_action( 'wp_footer', [ $this, 'processAjaxRequest' ], 99999 );
				}
			}
		}
	}

	/**
	 * Callback for the wp_enqueue_scripts hook: Enqueue scripts for the front-end
	 *
	 * @acces public
	 *
	 * @uses wp_enqueue_script()
	 *
	 * @since 0.1
	 */
	public function setScriptsFrontend()
	{
		global $wp_scripts;

		$exclude =  [];
		$uriAssetsJs = $this->globals->get( 'wfJsUri' );
		$pluginPath = $this->globals->get( 'pluginPath' );
		$jsUri = $this->globals->get( 'jsUri' );
		$isScriptDebug = WpieMiscHelper::isScriptDebug();
		$ext = ( $isScriptDebug ) ? '.js' : '.min.js';

		// make sure jQuery is enqueued
		wp_enqueue_script( 'jquery' );
		// enqueu common functions
		wp_enqueue_script( 'wpie-common',  $uriAssetsJs . '/wpie-common' . $ext, [], $this->globals->get( 'versionWf' ), true );

		if( file_exists( $pluginPath . '/assets/js/global' . $ext ) ) {
			wp_enqueue_script( $this->globals->nameSpace . '-global', $jsUri . '/global' . $ext, [ 'jquery', 'wpie-common' ], $this->globals->get('version') );
		}

		try {
			/**
			 * @todo description for hook activate_{$nameSpace}_scripts_frontend
			 */
			do_action( $this->globals->nameSpace . '_scripts_frontend', $wp_scripts, apply_filters( $this->globals->nameSpace . '_exclude_scripts_frontend', $exclude ), $isScriptDebug );
		} catch ( WpieExceptionInterface $e ) {
			WpieExceptionLogger::log( $e, true );
		}
	}

	/**
	 * Callback for the wp_head hook: setup global JavaScript parameters for the frontend
	 *
	 * @access public
	 *
	 * @uses json_encode() to safely create a JavaScript array
	 *
	 * @since 1.0
	 */
	public function printScriptsFrontendVars()
	{
		// WordPress installation URI
		$wpuri = get_bloginfo( 'wpurl' );
		// is multisite or not
		$isMs = WpieMultisiteHelper::isMs();

		// build array with vars
		$wpieVars = [];
		$wpieVars['ns'] = $this->globals->nameSpace;
		$wpieVars['nonce'] = wp_create_nonce( $this->globals->nameSpace . '-action' );
		$wpieVars['wpurl'] = $wpuri;
		$wpieVars['domain'] = WpieMiscHelper::getHostWithoutSubdomain( $wpuri );
		$wpieVars['ajaxurl'] = admin_url( 'admin-ajax.php' );
		$wpieVars['referer'] = wp_get_referer();
		$wpieVars['currenturl'] = ( $isMs ) ? WpieMultisiteHelper::getCurrentUri() : home_url( add_query_arg( NULL, NULL ) );
		$wpieVars['isms'] = ( $isMs ) ? true : false;
		$wpieVars['mspath'] = ( $isMs ) ? WpieMultisiteHelper::getBlogDetail( 'path' ) : '/' ;
		$wpieVars['ssl'] = is_ssl();

		// fix for install with a sub folder setup
		// @since 1.1.8
		if( false !== strpos( $wpieVars['currenturl'] , '/' ) && !$isMs ) {
			$wpieVars['currenturl'] = ( is_ssl() ? 'https://' : 'http://' ) . $_SERVER['HTTP_HOST'] . $_SERVER['REQUEST_URI'];
		}

		try {
			/**
			 * @todo description for hook {$nameSpace}_script_frontend_vars
			 */
			$wpieVars = apply_filters( $this->globals->nameSpace . '_script_frontend_vars' , $wpieVars );
		} catch( WpieExceptionInterface $e ) {
			WpieExceptionLogger::log( $e, true );
		}
		?>
		<script type='text/javascript'>
		/* <![CDATA[ */
		var <?php echo $this->globals->get( 'jsNamespace' ) ?> = <?php echo json_encode($wpieVars) ?>;
		/* ]]> */
		</script>
		<?php
	}

	/**
	 * Callback for the wp_head hook
	 *
	 * @access public
	 *
	 * @uses [YOUR_PLUGIN_NAMESPACE]_print_scripts_frontend to let WeePie Framework Plugins print styles for in the frontend head
	 *
	 * @since 0.1
	 */
	public function printScriptsFrontend()
	{
		try {
			/**
			 * @todo description for hook {$nameSpace}_print_scripts_frontend
			 */
			do_action( $this->globals->nameSpace . '_print_scripts_frontend' );
		} catch ( WpieExceptionInterface $e ) {
			WpieExceptionLogger::log( $e, true );
		}
	}


	/**
	 * Callback for the wp_enqueue_scripts hook
	 *
	 * @acces public
	 *
	 * @since 0.1
	 */
	public function setStylesFrontend()
	{
		global $wp_styles;

		$isScriptDebug = WpieMiscHelper::isScriptDebug();

		try {
			/**
			 * @todo description for hook {$nameSpace}_styles_frontend
			 */
		    do_action( $this->globals->nameSpace . '_styles_frontend', $wp_styles, apply_filters( $this->globals->nameSpace . '_exclude_styles_frontend', [] ), $isScriptDebug );
		} catch ( WpieExceptionInterface $e ) {
			WpieExceptionLogger::log( $e, true );
		}
	}

	/**
	 * Callback for the template_redirect hook: render front-end templates
	 *
	 * @access public
	 *
	 * @since 0.1
	 */
	public function renderTemplate()
	{
		try {
			/**
			 * @todo description for hook {$nameSpace}_render_templates
			 */
			do_action( $this->globals->nameSpace . '_render_templates' );
		} catch ( WpieExceptionInterface $e ) {
			WpieExceptionLogger::log( $e, true );
		}
	}
}